# --------------------
package Maps;
# --------------------
use Mojo::Base 'Mojolicious';

use Mojo::ByteStream 'b';
use Mojo::Cache;
use Mojo::Collection 'c';
use Mojo::File 'path';
use Mojo::Loader 'data_section';
use Mojolicious::Plugin::Markdown;
use Mojolicious::Plugin::FormFields;
use Mojolicious::Plugin::Token;
use Mojo::SQLite;

# use Env qw(DIGILIB_BASE);
use DDP;

our $VERSION = '0.2';

*_f = \&CORE::sprintf;

# runs once at server start
sub startup {
  my $self = shift;
  my $app = $self->app;

  # logs to STDERR
  $app->log->path(undef) if $app->mode eq 'development';

  # plugins
  $self->plugin('Config');
  $self->plugin('FormFields');
  $self->plugin('Markdown');
  $self->plugin('Token');

  my $config = $self->app->config;

  $config->{scaler_url} = $config->{digilib_base} . '/' . $config->{digilib_scaler};
  $config->{base_url} = $config->{digilib_base} . '/' . $config->{digilib_jquery};

  # cache

  $app->attr(_cache => sub { Mojo::Cache->new });

  # helpers

  $self->helper(cache_control => sub {
    my ($c, $age) = @_;
    $age ||= 3600;
    return $c->res->headers->cache_control("max-age=$age, must-revalidate");
    });

  $self->helper(datafile => sub {
    my ($c, $name, $pkg) = @_;
    $pkg ||= __PACKAGE__;
    return $config->{datafile}->{$name} ||= data_section($pkg, $name)
      || die("no data_section $pkg::$name");
    });

  $self->helper(xmlfile => sub {
    my ($c, $name) = @_;
    my $cached = $app->_cache->get($name);
    # return $cached if defined $cached;
    my $xml = b(path($config->{xmlpath}, "$name.xml")->slurp)->decode;
    $xml =~ s/^.+$//; # remove xml preamble
    $app->_cache->set($name => $xml);
    return $xml;
    });

  $self->helper(db => sub {
    my ($c, $key) = @_;
    my $dbname = $config->{db};
    state $db = Mojo::SQLite->new("sqlite:$dbname")->db;
    return $db;
    });

  $self->helper(digilib => sub {
    my ($c, $data) = @_;
    my $bbox = $data->{bbox};
    my $fac = $config->{digilib_factor} || 3;
    my $w = $bbox->{xmax} - $bbox->{xmin};
    my $h = $bbox->{ymax} - $bbox->{ymin};
    my $x = $bbox->{xmin} - $w * $fac;
    my $y = $bbox->{ymin} - $h;
    $x = '0.0' if $x < 0;
    $y = '0.0' if $y < 0;
    # TODO: config host name (automatically?), config factor
    return _f('/map?id=%s&fn=%s&mo=lpic&wx=%s&wy=%s&ww=%s&wh=%s',
      @$data{qw(id fn)}, $x, $y, $w * $fac * $fac, $h * $fac)
    });

  $self->helper(groups => sub {
    my ($c, $fn) = @_;
    return $c->ERROR('groups helper: no param "fn"') unless $fn;
    my $cached = $app->_cache->get($fn);
    return $cached if defined $cached;
    my $sql = $self->datafile('groups.sql');
    my $groups = $c->db->query($sql, $fn)->hashes;
    my $byid = $groups->reduce( sub { $a->{$b->{id}} = $b; return $a }, {});
    my $ids =  $groups->map( sub { return $_->{id} });
    $cached = { list => $groups, byid => $byid, ids => $ids };
    $app->_cache->set($fn => $cached);
    return $cached;
    });

  $self->helper(groupname => sub {
    my ($c, $fn, $id) = @_;
    my $groups = $self->groups($fn)->{byid};
    my $group = $groups->{$id} || $groups->{1};
    return $group->{name};
    });

  $self->helper(osmlink => sub {
    my ($c, $res) = @_;
    my $conf = $config->{osm};
    my ($fac, $temp) = ($conf->{factor}, $conf->{template});
    my ($lat, $lon) = ($res->{lat}, $res->{lon});
    return _f($temp, $lat, $lon, $fac, $lat, $lon);
    });

  $self->helper(mapinfo => sub {
    # TODO: check for existing image file (fn) ?
    my ($c, $fn) = @_;
    $fn ||= $c->session('fn');
    return {} unless $fn;
    $c->db->query('INSERT OR IGNORE INTO map (fn) VALUES (?)', $fn);
    return $c->db->query(
      $self->datafile('map.sql'),
      $fn,
      )->hash;
    });

  $self->helper(DEBUG => sub { shift->app->log->debug(@_) });
  $self->helper(ERROR => sub { shift->app->log->error(@_) });

  $app->secrets($config->{secrets});

  # router
  my $r = $self->routes;

  # Normal route to controller
  $r->get('/')->to('pages#index');
  $r->post('/authenticate')->to('pages#authenticate');

  $r->get('/map')->to('pages#map');
  $r->get('/napoli')->to('pages#napoli');
  $r->get('/guide/:name')->to('pages#transcription');

  # toponyms
  $r->get('/top')->to('top#all');
  $r->get('/top/:id')->to('top#item');
  $r->get('/top/:id/:format')->to('top#json');

  # handle token / CORS
  $r->post('/token')->to('pages#token');
  $r->options('/map/*')->to('pages#options');

  my $valid = $r->under('/map')->to('pages#validate');

  # search
  $valid->any('/search')->to('db#search');

  # store
  $valid->post('/annotations')->to('db#create');
  $valid->get('/annotations/:id')->to('db#get');
  $valid->put('/annotations/:id')->to('db#update');
  $valid->delete('/annotations/:id')->to('db#del');

  # item is constrained
  # $r->get('/db/:item' => [item => $config->{api}] )->to('db#item');
  return;
}

__PACKAGE__

__DATA__
@@groups.sql
SELECT g.*, COUNT(i.id) AS count, GROUP_CONCAT(i.id) AS ids
FROM groups g LEFT JOIN item i
ON g.ID = i.ref_group AND i.fn = ? AND i.state > 0
GROUP BY g.id
ORDER BY g.name

@@map.sql
SELECT m.*, COUNT(i.ID) AS items
FROM map m LEFT JOIN item i
ON m.ID = i.ref_map AND m.fn = ? AND i.state > 0
GROUP BY m.id
ORDER BY m.line

