# --------------------
package Maps::Controller::Db;
# --------------------
use Mojo::Base 'Maps::Controller';
use Mojo::JSON 'encode_json';
use Mojo::Loader 'data_section';
use Mojo::Template;
use Mojo::URL;

use Const::Fast;
use List::Util qw(max min);
use DDP;

our $VERSION = '0.2';

const my $EMPTY => q{};
const my $SPACE => q{ };
const my $NO_CONTENT => 204;

#	DBI TraceLevel 	Description
	#	0   Trace disabled.
	#	1   Trace top-level DBI method calls returning with results or errors.
	#	2   As above, adding tracing of top-level method entry with parameters.
	#	3   As above, adding some high-level information from the driver
	#	    and some internal information from the DBI.
	#	4   As above, adding more detailed information from the driver.
	#	    This is the first level to trace all the rows being fetched.
	#	5-15 As above but with more and more internal information

#	$db->dbh->trace( 1, 'dbi_trace.log' );

#	$ENV{DBI_TRACE} = "1=dbi_trace.log";

*_f = \&CORE::sprintf;

my $_allFields = [qw(
  id
  ref_map ref_group ref_top fmid qid fn name nr legend desc type shape bbox tags comment
  state user created
  )];

my $_updateFields = [ @$_allFields[1..15] ];
my $_countAll = @$_allFields;
my $_countUpdate = @$_updateFields;
my $_insertFieldnames = join (q{, }, $_allFields);
my $_insertPlaceholders  = join(q{, }, ('?') x ($_countAll - 1));
my $_updatePlaceholders = join (q{, }, map { "$_ = ?" } $_countUpdate);

sub _renderCached {
  my ($c, $result) = @_;
  $c->cache_control->render(json => $result);
}

# calculate the bounding box of the coordinates

sub _bbox {
  my ($c, $coords) = @_;
  my @x = map { $_->[0] } @$coords;
  my @y = map { $_->[1] } @$coords;
  return [min(@x), max(@x), min(@y), max(@y)];
}

# return a shapes data structure

sub _shapes {
  my ($c, $shape) = @_;
  return [{
    type => 'polygon',
    geometry => {
      type => 'Polygon',
      units => 'fraction',
      coordinates => $shape,
      },
    }];
}

# add CSS, shapes, permissions and textfield to $json (= the db result)

sub _jsonCreate {
  my ($c, $json) = @_;
  my $fn = $c->param('fn') || Mojo::URL->new($json->{uri})->query->param('fn');
  my $template = Mojo::Template->new(vars => 1);
  my $textfield = $c->datafile(textfield => __PACKAGE__);
  my $groupname = $c->groupname($fn, $json->{ref_group});
  my $nr = $json->{nr};
  $json->{cssclass} = "map-$groupname";
  $json->{groupname} = $groupname;
  $json->{label} = sprintf('%s, %s%s', $json->{name}, $groupname, $nr ? " [$nr]" : '', );
  $json->{permissions} ||= $c->config->{permissions};
  $json->{shapes} = $c->_shapes($json->{shape});
  $json->{text} = $template->render($textfield, $json);
  return $json;
}

# return an array of field contents extracted from $json

sub _jsonValues {
  my ($c, $json) = @_;
  my $fn = Mojo::URL->new($json->{uri})->query->param('fn');
  my $map = $json->{ref_map} || $c->mapinfo($fn)->{ID};
  my $top = $json->{ref_top};
  my $group = $json->{ref_group} || 1;
  my $fields = [ @$json{qw(name nr legend desc)} ];
  my $fmid = $json->{fmid} || 0;
  my $qid = $json->{qid} || '';
  my $shape = $json->{shapes}->[0];
  my $type = $shape->{type};
  my $points = $shape->{geometry}->{coordinates};
  my $bbox = $c->_bbox($points);
  my $tags = join $SPACE, @{$json->{tags}};
  my $comment = $json->{comment} || '';
  return [$map, $group, $top, $fmid, $qid, @$fields, $type, $points, @$bbox, $tags, $comment];
}

# get user information for name

sub user {
  my $c = shift;
  my $name = $c->param('name') || $EMPTY;
  my $result = $c->db->query('SELECT * FROM user WHERE name = ?', $name)->hash;
  return $c->_renderCached($result);
}

# get groups, objects and map metadata for an image filename

sub search {
  my $c = shift;
  my $fn = $c->param('fn');
  my $db = $c->db;
  my $user = $c->stash('user');
  my $map = $c->mapinfo($fn);
  my $groups = $c->groups($fn)->{list};
  my $sql = 'SELECT * FROM item i WHERE i.state > 0 AND i.fn = ? ORDER BY type, name';
  my $items = $c->db
  	->query($sql, $fn)
  	->expand(json => [qw(shape bbox)])
  	->hashes
  	->map(sub {
  	   $c->_jsonCreate($_)
  	   });
  # return complete results (map, groups and items) for image (fn)
  my $result = { map => $map, groups => $groups, rows => $items, total => $items->size };
  $c->DEBUG("DB search - fn: $fn, user: $user, annotation count: " . $items->size);
  return $c->render(json => $result);
}

# create an annotation from the web (in table 'item')

sub create {
  my $c = shift;
  my $user = $c->stash('user');
  my $json = $c->req->json;
  my $fields = $c->_jsonValues($json);
  my $sql = _f(q{INSERT into item (%s) VALUES (%s, datetime('now','localtime'))}, $_insertFieldnames, $_insertPlaceholders);
  my $id = $c->db->query(undef, @$fields, $user)->last_insert_id;
  # $c->db->query('UPDATE item SET shape = ? WHERE ID = ?', {shape => $json->{shape}}, $id);
  my $fn = $fields->[5];
  $c->DEBUG("DB create - fn: $fn, annotation: $id ($user)");
  $c->render(json => $c->_jsonCreate($json));
}

# update an annotation from the web (into table 'item')

sub update {
  my $c = shift;
  my $id = $c->stash('id');
  my $user = $c->stash('user');
  my $json = $c->req->json;
  my $fields = $c->_jsonValues($json);
  my $sql = _f('UPDATE item SET %s WHERE id = ?', $_updatePlaceholders);
  $c->db->query($sql, @$fields, $id);
  $c->DEBUG("DB update - annotation: $id ($user)");
  $c->render(json => $c->_jsonCreate($json));
}

# retrieve a single annotation (from table 'item')

sub get {
  my $c = shift;
  my $id = $c->stash('id');
  my $user = $c->stash('user');
  my $res = $c->db->query('SELECT * FROM item WHERE ID = ?', $id)->hash;
  $c->DEBUG("DB get - annotation: $id ($user)");
  $c->render(json => $c->_jsonCreate($res));
}

# delete a single annotation (from table 'item')

sub del {
  my $c = shift;
  my $id = $c->stash('id');
  my $user = $c->stash('user');
  $c->db->query('UPDATE item SET state = 0 WHERE ID = ?', $id);
  $c->DEBUG("DB delete - annotation: $id ($user)");
	return $c->render(text => $EMPTY, status => $NO_CONTENT);
}

__PACKAGE__

# construct markdown for annotator balloon

__DATA__

@@textfield
__<%= qq|[$nr] | if $nr %><%= $name %>__

<%= qq|_legend:_ "$legend"| if $legend %>

<%= qq|_group:_ `$groupname`| %>  <%= $tags ? qq|_tags:_ `$tags`| : '' %>
<%= $desc ? qq|_desc:_ ${desc}| : '' %>

<%= sprintf(qq{[gazetteer](http://localhost:3000/top/%s)}, $ref_top) if $ref_top %>

<%= sprintf(qq{[zuccaro](http://zuccaro/biblhertz.it/building/%s)}, $fmid) if $fmid %>

<%= sprintf(qq{[wikidata](https://www.wikidata.org/wiki/Q%s)}, $qid) if $qid %>

