#!/usr/bin/env perl
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Mojo::JSON::Pointer;
use Mojo::Collection;

use DDP;
use Const::Fast;

our $VERSION = '0.1';

const my $OK => 200;
const my $NO_CONTENT => 204;
const my $MOVED_PERMANENTLY => 301;
const my $FOUND => 302;
const my $SEE_OTHER => 303;
const my $BAD_REQUEST => 400;
const my $UNAUTHORIZED => 401;
const my $FORBIDDEN => 403;
const my $NOT_FOUND => 404;
const my $INTERNAL_SERVER_ERROR => 500;

const my $DIGILIB => 'http://localhost:8888';
const my $URI => 'http://localhost:8011/digilib/jquery/map.html';
const my $FN => 'Azr373a.jpg';
const my $ZA => '1x1@0,0';
const my $ORIGIN => 'localhost';

$ENV{DIGILIB_BASE} = $DIGILIB;

my $t = Test::Mojo->new('Maps');
my $app = $t->app;
my $config = $app->config;
my $commands = $app->commands;
my $jconf = _jp($config);

sub _jp { Mojo::JSON::Pointer->new(shift) };
sub _jp_tx { _jp($t->tx->res->json) };

# create a test database
$config->{db} = 'db/test.sqlite';
my $db = $app->db;

# basic db structure
like $db->sqlite->dsn, qr{test.sqlite$}, 'db name ok';
ok $commands->run('create'), 'db created';
ok $commands->run('import'), 'db imported';
ok 4 == scalar(grep { /"map|item|groups|user"/ } @{$db->tables}), 'all tables exist';
ok 2 == $db->query('select fn from map')->arrays->size, 'number of maps ok';
ok 'none' eq $db->query('select role from user where name = ?', 'import')->array->[0], 'role ok';
ok 'admin' eq $db->query('select role from user where name = ?', 'admin')->array->[0], 'admin role ok';
my $groupcount = $db->query('select name, color from groups')->arrays->size;
ok 1 < $groupcount, 'number of groups > 1';
ok 'HASH' eq ref $app->groups, 'groups helper returns hash';
ok $groupcount == scalar keys %{$config->{groups}}, 'config groups ok';
ok $DIGILIB eq $config->{digilib_base}, 'digilib base';
like $config->{scaler_url}, qr{^$DIGILIB}, 'scaler url';
like $config->{base_url}, qr{^$DIGILIB}, 'base url - server';
like $config->{base_url}, qr{jquery/}, 'base url - jquery';

# Authentication
# $db->query('delete from item');
$t->reset_session;
$t->post_ok('/token')->status_is($UNAUTHORIZED)->content_is('Unauthorized');
$t->post_ok('/token', form => {user => 'admin', password => 'tadoadsty'})->status_is($OK);
my $token = $t->tx->res->text;
my $claims = $app->validateToken($token);
ok $token eq $app->issueToken('admin'), 'token ok';
ok $claims, 'token valid';
ok 'admin' eq $claims->{userId}, 'token payload ok';

# Cookie (not tested)
$t->ua->cookie_jar->add(Mojo::Cookie::Response->new(
  name   => 'digilib:fn:Azr373a.jpg:pn:1',
  value  => "annotationUser=admin&annotationToken=$token",
  domain => 'localhost',
  path   => '/digilib/jquery/',
  ));
my $cookies = $t->ua->cookie_jar->all;

# search for annotations
my $headers = {
	'x-annotator-auth-token' => $token,
	'origin' => $ORIGIN,
	};
$t->get_ok("/map/search?fn=$FN&uri=$URI&za=$ZA")->status_is($UNAUTHORIZED);
$t->get_ok("/map/search?fn=$FN&uri=$URI&za=$ZA", $headers)->status_is($OK);

# CORS response
my $h = $t->tx->res->headers;
ok 'true' eq $h->header('access-control-allow-credentials'), 'credentials header ok';
ok $ORIGIN eq $h->header('access-control-allow-origin'), 'origin header ok';

# annotations content
my $json = $t->tx->res->json;
$t->json_has('/groups');
$t->json_is('/map/fn', $FN);
$t->json_has('/rows');
$t->json_is('/total', 0);

ok $groupcount == scalar @{$json->{groups}}, "$groupcount groups";
ok $json->{total} == scalar @{$json->{rows}}, 'rows count ok';

# post new annotation
my $group = 8;
my $groupname = $jconf->get("/groups/$group/name");
my $annotation =  {
  desc => 'Beschreibung',
  fmid => '999',
  group => $group,
  legend => 'Legende',
  name => 'Normname',
  nr => '1234',
  ranges => [],
  shapes => [{
    type => 'polygon',
    geometry => {
      coordinates => [ [0.34, 0.72], [0.41, 0.74], [0.39, 0.80], [0.33, 0.77] ],
      units => 'fraction',
      },
    }],
  tags => ['tag1', 'tag2'],
  uri => 'http://localhost:8011/jquery/map.html?fn=Azr373a.jpg',
  };

$t->options_ok('/map/annotations', {origin => $ORIGIN }, json => $annotation)->status_is($OK);

# CORS response
$h = $t->tx->res->headers;
ok 'true' eq $h->header('access-control-allow-credentials'), 'credentials header ok';
ok $ORIGIN eq $h->header('access-control-allow-origin'), 'origin header ok';

# post annotation
my $re_text = qr{^_[^_]+_\n\nnr \d+: .*?\n\nfmid: \d+\n\ngroup: \w+\n\n.*$}; # change here
$t->post_ok('/map/annotations', json => $annotation)->status_is($UNAUTHORIZED);
$t->post_ok('/map/annotations', $headers, json => $annotation)->status_is($OK);
$t->json_has('/id');
$t->json_has('/shapes/0/geometry/coordinates');
$t->json_hasnt('/shapes/1');
$t->json_hasnt('/fn'); # no fn field in json
$t->json_like('/text', $re_text);

my $items = $db->query('select * from item')->hashes;
ok 1 == $items->size, '1 item';

# post second annotation
$annotation->{uri} =  'http://localhost:8011/jquery/map.html?fn=xyz';
$t->post_ok('/map/annotations', $headers, json => $annotation)->status_is($OK);
$t->json_hasnt('/fn'); # no fn field in json

$items = $db->query('select * from item')->hashes;
ok 2 == $items->size, '2 items';

# search all annotations for uri
$t->get_ok("/map/search?fn=$FN&uri=$URI&za=$ZA", $headers)->status_is($OK);
$t->json_is('/total', 1);
$t->json_has('/rows/0');
$t->json_has('/rows/0/id');
$t->json_has('/rows/0/text');
$t->json_has('/rows/0/cssclass');
$t->json_is('/rows/0/cssclass', "map-$groupname");
$t->json_hasnt('/rows/1');

# get one annotation
my $id = _jp_tx->get('/rows/0/id');
$t->get_ok('/map/annotations/', $headers)->status_is($NOT_FOUND);
$t->get_ok("/map/annotations/$id")->status_is($UNAUTHORIZED);
$t->get_ok("/map/annotations/$id", $headers)->status_is($OK);
# my $jp = _jp_tx;
$t->json_is('/id', $id);
$t->json_is('/name', 'Normname');
$t->json_like('/text', qr{^_Normname_});

# update one annotation
$annotation->{name} = 'Noname';
$t->put_ok('/map/annotations', json => $annotation)->status_is($NOT_FOUND);
$t->put_ok("/map/annotations/$id", json => $annotation)->status_is($UNAUTHORIZED);
$t->put_ok("/map/annotations/$id", $headers, json => $annotation)->status_is($OK);
$t->json_hasnt('/fn'); # no fn field in json
$t->json_is('/id', $id);
$t->json_is('/name', 'Noname');
$t->json_like('/text', qr{^_Noname_});

# delete one annotation
$t->delete_ok('/map/annotations')->status_is($NOT_FOUND);
$t->delete_ok("/map/annotations/$id")->status_is($UNAUTHORIZED);
$t->delete_ok("/map/annotations/$id", $headers)->status_is($NO_CONTENT);

# search all annotations for uri
$t->get_ok("/map/search?fn=$FN&uri=$URI&za=$ZA", $headers)->status_is($OK);
$t->json_is('/total', 0);
$t->json_has('/rows');
$t->json_hasnt('/rows/0');

# check db for deleted item
$items = $db->query('select id from item')->hashes;
ok 2 == $items->size, '2 items';
$items = $db->query('select id from item where state = 0')->hashes;
ok 1 == $items->size, '1 item deleted';

# get map page
my $base = $config->{base_url};
like $base, qr{^http:\/\/}, 'base url is absolute';
like $base, qr{^$config->{digilib_base}}, 'base url is digilib';
$t->get_ok("/map?fn=$FN")->status_is($OK)
  ->element_exists("base[href=$base]", 'base tag')
  ->element_exists('div#digilib')
  ->element_exists('script[src=jquery.js]')
  ->element_exists('script[src=jquery.digilib.js]')
  ->element_exists('script[src=annotator-dl.js]')
  ->element_exists('script[src=jquery.digilib.annotator.js]')
  ->element_exists('script[src=showdown.js]')
  ->element_exists('link[href=jquery.digilib.css')
  ->element_exists('link[href=annotator.min.css')
  ->element_exists('style#maps-local-css')
  ->element_exists('style#maps-groups-css')
  ->element_exists('script#maps-local-js')
  ->element_exists('div#map-groups')
  ->element_exists('input#cb-group-all');
my $dom = $t->tx->res->dom;
my $inputs = $dom->find('#map-groups input')->size;
ok $groupcount == $inputs - 1, 'group checkbox count';

done_testing();

